from .ipython_blocking import CaptureExecution
from .ipython_magic import CaptureMagic, load_ipython_extensions

__version__ = "0.3.0"

try:
    # This won't load when get_ipython() isn't defined, e.g. in default python shell.
    # In an IPython/Jupyter environment, I want the %block and %blockrun magic
    # to be available by default
    load_ipython_extensions()
except:
    pass



import sys

import tornado.queues
from nbclient.util import just_run

### General context manager usage:

# ctx = CaptureExecution()
# with ctx:
#     while True:
#         if <something -- widget change or validation function returns True>:
#             break
#         ctx.step()


class CaptureExecution:
    "A context manager to capture execute_request events then either replay or disregard them after exiting the manager"

    def __init__(self, replay=True):
        self.captured_events = []
        self._replay = replay
        self.shell = get_ipython()
        self.kernel = self.shell.kernel

    def step(self):
        try:
            just_run(self.kernel.do_one_iteration())
        except tornado.queues.QueueEmpty:
            pass

    def capture_event(self, stream, ident, parent):
        "A 'capture' function to register instead of the default execute_request handling"
        self.captured_events.append((stream, ident, parent))

    def start_capturing(self):
        "Overwrite the kernel shell handler to capture instead of executing new cell-execution requests"
        self.kernel.shell_handlers["execute_request"] = self.capture_event

    def stop_capturing(self):
        "revert the kernel shell handler to the default execute_request behavior"
        self.kernel.shell_handlers["execute_request"] = self.kernel.execute_request

    def replay_captured_events(self):
        "Called at end of context -- replays all captured events once the default execution handler is in place"
        # need to flush before replaying so messages show up in current cell not replay cells
        sys.stdout.flush()
        sys.stderr.flush()
        for stream, ident, parent in self.captured_events:
            # Using kernel.set_parent is the key to getting the output of the replayed events
            # to show up in the cells that were captured instead of the current cell
            self.kernel.set_parent(ident, parent)
            just_run(self.kernel.execute_request(stream, ident, parent))

    def __enter__(self):
        self.start_capturing()
        # increment execution count to avoid collision error
        self.shell.execution_count += 1

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            # let the error propogate up, such as a keyboard interrupt while capturing cell execution
            return False
        self.stop_capturing()
        if self._replay:
            self.replay_captured_events()



import time
import types

import ipywidgets as widgets
from IPython.core import magic_arguments
from IPython.core.magic import Magics, line_magic, magics_class
from IPython.display import Javascript, display

from .ipython_blocking import CaptureExecution


def run_all_below():
    js = "Jupyter.notebook.select_next().execute_cells_below()"
    display(Javascript(js))


@magics_class
class CaptureMagic(Magics):
    def capture(self, breaking_func, timeout=None, replay=True):
        if timeout:
            timeout = int(timeout)
        start = time.time()
        ctx = CaptureExecution(replay=replay)
        with ctx:
            while True:
                if breaking_func():
                    break
                if timeout:
                    if (time.time() - start) >= timeout:
                        break
                ctx.step()

    @line_magic
    @magic_arguments.magic_arguments()
    @magic_arguments.argument(
        "break_value",
        help="Widget object or function that defines a break from the blocking context",
    )
    @magic_arguments.argument(
        "-t", "--timeout", default=None, help="Timeout in seconds to stop capturing"
    )
    def block(self, line):
        line = line.strip()
        args = magic_arguments.parse_argstring(self.block, line)

        obj = get_ipython().user_ns[args.break_value]
        ### Support the following options for a break value:
        ### 1) a callable function that will break when the function returns True
        ### 2) a ValueWidget, which will break when the value changes
        ### 3) a ButtonWidget, which will break when it is clicked
        if isinstance(obj, (types.FunctionType, types.MethodType)):
            func = obj
        elif isinstance(obj, widgets.ValueWidget):
            starting_value = obj.value
            func = lambda: obj.value != starting_value
        elif isinstance(obj, widgets.Button):
            obj._has_been_clicked = False

            def handler(w):
                w._has_been_clicked = True

            obj.on_click(handler)
            func = lambda: obj._has_been_clicked
        else:
            raise Exception(
                "The positional argument to %block should be a ValueWidget, Button, or a function/method"
            )
        return self.capture(func, args.timeout)

    @line_magic
    @magic_arguments.magic_arguments()
    @magic_arguments.argument(
        "button_widget", help="A Button Widget to block-and-run-all-below on"
    )
    @magic_arguments.argument(
        "-t", "--timeout", default=None, help="Timeout in seconds to stop capturing"
    )
    def blockrun(self, line):
        line = line.strip()
        args = magic_arguments.parse_argstring(self.blockrun, line)

        obj = get_ipython().user_ns[args.button_widget]
        ### When a cell -> run all is executed, add a _has_been_clicked attribute to the widget
        ###   and a handler for a "run all cells below"

        ### If _has_been_clicked is True, do nothing (let the rest of the notebook run)
        ### Otherwise block and replay=False
        if not hasattr(obj, "_has_been_clicked"):
            obj._has_been_clicked = False

            def handler(w):
                w._has_been_clicked = True
                run_all_below()

            obj.on_click(handler)

        if not getattr(obj, "_has_been_clicked"):
            return self.capture(
                lambda: obj._has_been_clicked, timeout=args.timeout, replay=False
            )


def load_ipython_extensions():
    get_ipython().register_magics(CaptureMagic)



from setuptools import setup, find_packages

setup(name='ipython_blocking',
      version='0.3.0',
      
      author='Matt Kafonek',
      author_email='kafonek@gmail.com',
      url='https://github.com/kafonek/ipython_blocking',
            
      description='Context manager for blocking cell execution within a Jupyter notebook',
      long_description=open('README.md', encoding='utf-8').read(),
      long_description_content_type = 'text/markdown',
      
      packages=find_packages(),
      install_requires=['IPython', 'ipywidgets', 'ipykernel>=6', 'nbclient', 'tornado'],
      
      classifiers = ['Framework :: IPython',
                     'Programming Language :: Python',
                     'Intended Audience :: Developers',
                     'Development Status :: 3 - Alpha',
                     ],
      license='BSD')
